MODULE c_ariaux;

{
  Hilfsroutinen fuer die beiden Module C_ARI und CI_ARI.   Stand:  19.12.91

  Die Prozedur  produkt  berechnet das kurze Skalarprodukt a*b + c*d,
  wobei moegliche Over- und Underflows durch Skalieren und Setzen ent-
  sprechender Flags vermieden wird. Das Ergebnis wird als Dotprecision
  geliefert.

  Die Funktion  quotient  liefert den Quotienten zweier Dotprecision,
  die mit der Prozedur  produkt  berechnet wurden. Eventuell gesetzte
  Over-, Underflowflags werden beruecksichtigt.


  !!!  Um das Modul fuer den XSC-C-Precompiler zu verwenden, muessen  !!!
  !!!  alle Kommentare
  !!!                ($on)     nach   $on
  !!!                ($off)    nach   $off
  !!!                ($XSC-C   nach   ($XSC-C)
  !!!                XSC-C$)   nach   (XSC-C$)
  !!!  geaendert werden. Die andere Richtung gilt sinngemaess.
  !!!  Dabei stehen hierin die runden Klammern fuer die eigentlichen Kommentarklammern
  !!!  ( welche innerhalb dieses Kommentares nicht verwendet werden koennen ).
  !!!  Achtung: Keine geschachtelten Kommentare erzeugen !!!

Aenderungen:

R.Lohner   --.--.--   Urversion mit DORTPRECISION als Ausgabe von produkt und
                      Eingabe von quotient.
R.Lohner   09.12.91   Umstellung auf Staggered p1+p2 bei produkt Ausgabe und
                      quotient Eingabe. Es wird Typ Interval verwendet, aber
                      nicht Modul i_ari.
R.Lohner   23.01.92   in XSC-C Version: maxreal-Definition aus Modul x_real.
}


$off

{ Konstanten fuer Dezimalarithmetik (ATARI) : }

GLOBAL CONST
             maxreal  = 9.999999999999E99;
             mantlen  = 53;    { Mantissenlaenge }
             minexpo  = -98;   { Exponentenbereich = [-98,100] ! }
             maxexpo  = 100;
             minexpom = -71;   { = minexpo + 1 + 2*mantlen }
             maxexpo1 =  99;   { = maxexpo - 1 }

$on

{ Konstanten fuer IEEE-Arithmetik (Bedeutung wie bei ATARI) : }

{$XSC-C}

use x_real;

GLOBAL CONST
	{    maxreal  = 1.7976931348623157e308;  } { in x_real definiert }
	     mantlen  = 53;
	     minexpo  = -1021;
	     maxexpo  = 1024;
	     minexpom = -914;
	     maxexpo1 = 1023;
{XSC-C$}


GLOBAL PROCEDURE produkt( a,b,c,d : REAL; VAR overfl : INTEGER;
                           VAR p1 : REAL; VAR p2 : INTERVAL );

{ Berechnet a*b + c*d als Staggered-Groesse p1 + p2          }
{ Signalisiert ausserdem  Ueber-/Unterlauf :                 }
{       overfl = -1,0,1  ( Unterlauf/Nix/Ueberlauf )         }

var exa,exb,exc,exd : integer;  { Exponenten von a,b,c,d }
    dot             : dotprecision;
    inexact         : boolean;
begin
  overfl := 0;
  inexact:= false;

  exa:= expo(a);
  exb:= expo(b);
  exc:= expo(c);
  exd:= expo(d);
  if ( sign(a) = 0 ) or ( sign(b) = 0 ) then
     { a = b = 0 }
     if ( sign(c) = 0 ) or ( sign(d) = 0 ) then
        { a = b = c = d = 0 }
        dot := #(0)
     else
     begin
        { a = b = 0, c,d <> 0 }
        if exc+exd > maxexpo1 then
        begin
           { overflow ! }
           c:= comp( mant(c), exc-maxexpo );
           overfl:= 1;
        end else
        if  exc+exd < minexpom then 
        { minexpom, da Unterlauf auch durch Ausloeschung moeglich ist }
        begin
           { underflow ! }
           c:= comp( mant(c), exc+maxexpo );
           overfl:= -1;
        end;
        dot:= #( c*d );
     end
  else
     { a,b <> 0 }
     if ( sign(c) = 0 ) or ( sign(d) = 0 ) then
     begin
        { a,b <> 0, c = d = 0 }
        if exa+exb > maxexpo1 then
        begin
           { overflow ! }
           a:= comp( mant(a), exa-maxexpo );
           overfl:= 1;
        end else
        if  exa+exb < minexpom then
        { minexpom, da Unterlauf auch durch Ausloeschung moeglich ist }
        begin
           { underflow ! }
           a:= comp( mant(a), exa+maxexpo );
           overfl:= -1;
        end;
        dot:= #( a*b );
     end else
     begin
        { a,b,c,d <> 0 }
        if exa+exb > maxexpo1 then
        begin
           { overflow bei a*b }
           a:= comp( mant(a), exa-maxexpo );
           if exc>0 then c:= comp( mant(c), exc-maxexpo ) else
           if exd>0 then d:= comp( mant(d), exd-maxexpo ) else
           begin
              { underflow wegen Skalierung bei c*d }
              c:= 0;
              inexact:= true;
           end;
        end else
        if exc+exd > maxexpo1 then
        begin
           { overflow bei c*d }
           c:= comp( mant(c), exc-maxexpo );
           if exa>0 then a:= comp( mant(a), exa-maxexpo ) else
           if exb>0 then b:= comp( mant(b), exb-maxexpo ) else
           begin
              { underflow wegen Skalierung bei a*b }
              a:= 0;
              inexact:= true;
           end;
           overfl:= 1;
        end else
        if ( exa+exb < minexpom ) and ( exc+exd < minexpom ) then
        begin
           { underflow bei a*b und bei c*d }
           a:= comp( mant(a), exa+maxexpo );
           c:= comp( mant(c), exc+maxexpo );
           overfl:= -1;
        end;
        dot:= #( a*b + c*d );
     end;

     p1:= #*( dot );
     p2:= ##( dot - p1 );

     if inexact then
     begin
        p2.inf:= pred(p2.inf);
        p2.sup:= succ(p2.sup);
     end;
     
END;    { produkt }



GLOBAL FUNCTION quotient( z1 : REAL; z2 : INTERVAL; n1 : REAL; n2 : INTERVAL;
                          round, zoverfl, noverfl : INTEGER ) : REAL;

{ Berechnet den Quotienten  (z1+z2) / (n1+n2) mittels staggered     }
{ correction Technik.                                               }
{ Rundung round (-1,0,1) wird beruecksichtigt.                      }
{ zoverfl und noverfl werden fuer die Rueckskalierung beruecksich-  }
{ tigt ( wenn Zaehler-/Nenner- Over-/Underflow vorlag ).            }
{ Vom Nenner wird angenommen : n1+n2 > 0 !                          }

VAR q1,nlo1,nlo2,nup1,nup2,scale : REAL;
    q2,nh                        : INTERVAL;
    vorz,anz_scale               : INTEGER;
BEGIN
  vorz:= sign(z1);  { z2 ??? }

  IF ( zoverfl=-1 ) AND ( noverfl=1 ) THEN
     { Ergebnis liegt im Unterlaufbereich }
  CASE round OF
    -1 : IF vorz>=0 THEN q1:= 0.0 ELSE q1:= pred( 0.0 );
     0 : q1:= 0.0;
    +1 : IF vorz<=0 THEN q1:= 0.0 ELSE q1:= succ( 0.0 );
  END ELSE
  IF ( zoverfl=1 ) AND ( noverfl=-1 ) THEN
     { Ergebnis liegt im Ueberlaufbereich }
  CASE round OF
    -1 : IF vorz>=0 THEN q1:= maxreal ELSE q1:= -maxreal-maxreal; { Overflow }
     0 : q1:= maxreal+maxreal;                                    { Overflow }
    +1 : IF vorz<=0 THEN q1:=-maxreal ELSE q1:=  maxreal+maxreal; { Overflow }
  END ELSE
  BEGIN
    q1:= z1 /< n1;  { /< damit q2 >=0 wird }

    nh.inf:= n1 +< n2.inf;
    nh.sup:= n1 +> n2.sup;

    q2:= ##( z1 - q1*n1 + z2 - q1*n2 );

    case round of
    -1 :  q1:= q1 +< q2.inf/<nh.sup;
     0 :  q1:= q1 + (q2.inf+q2.sup)*0.5/n1;
    +1 :  q1:= q1 +> q2.sup/>nh.inf;
    end;
    
    { Rueckskalierung, falls Zaehler/Nenner - Over/Underflow : }

  {   Eigentlich wie folgt:

    q1:= comp( mant(q1), expo(q1) + (zoverfl-noverfl)*maxexpo );
  }
    { Da aber obiges bei zoverfl-noverfl < 0 unkontrollierten underflow }
    { erzeugen kann ( mit Fehlermeldung 'inexact' vom Laufzeitsystem ), }
    { wird gerichtet gerundet zurueck skaliert :                        }

    anz_scale:= zoverfl - noverfl;
    if anz_scale > 0 then scale:= comp( mant(1.0), expo(1.0)+maxexpo ) else
    if anz_scale < 0 then scale:= comp( mant(1.0), expo(1.0)-maxexpo ) 
                     else scale:= 1.0;
    anz_scale:= abs(anz_scale);
    case round of
    -1: begin
           q1:= q1 *< scale;
           if anz_scale=2 then q1:= q1 *< scale;
        end;
     0: begin
           q1:= q1 * scale;
           if anz_scale=2 then q1:= q1 * scale;
        end;
    +1: begin
           q1:= q1 *> scale;
           if anz_scale=2 then q1:= q1 *> scale;
        end;
    end;

    
  END;

  quotient:= q1;

END;


END.
