(*                        }
{*  PROGRAM:  psclist.p   }
{*  Listing generator of the PASCAL-XSC to C compiler.             }
{*  Merges pascal source lines with linenumbers and errormessages. }
{*  last modification:  17.06.91  10:40   }
{*  uli 17.6.91: errtext changed errtext.hlp, errmess changed to errmess.tmp }
{*)

PROGRAM PSCLIST (input, output);

(* ERZEUGT AUS DEM QUELLTEXT UND DER FEHLERMELDUNGEN-DATEI ERRMESS *)
(* EIN LISTING MIT FEHLERMELDUNGEN                                 *)

use iostd ;

CONST
  errtext1 = 'errtext.hlp'    ; { filename }
  errmess1 = 'errmess.tmp'    ; { filename or error message file }
  full_listing      = false   ; { false ==> exsc ; true ==> psclist }
  ordtab            =  9      ; { ord (tab) for ascii, adjust for ebcdic }
  subchar           = '@'     ; { substituted by additional string }

   infolimit        =    10 ;  { These constants must corr. with those }
  warnglimit        =   100 ;  { constants of the C-module error.h     }
  errorlimit        =  1000 ;
  checklimit        =  9000 ;
  fatallimit        = 30000 ;

  retchar           = '*';
  (* KENNZEICHEN FUER DEN AUFSETZPUNKT NACH GLOBALER FEHLERBEHANDLUNG *)
  errlistsize       = 500;       (* MAXIMALE ANZAHL DER FEHLER *)
  errtxtsize        = 30000;     (* interner Heap fuer strings *)
  (* GROESSE DES ARRAYS, DAS SAEMTLICHE FEHLERTEXTE ENTHAELT *)

   infoat           = ' Info at';
  warngat           = 'Warng at';
  errorat           = 'Error at';
  checkat           = 'Check at';
  fatalat           = 'Fatal at';

TYPE

  indextype       = 0 .. errlistsize;

VAR

 { source:             TEXT ; substitute for input }
 { listing:            TEXT;  substitute for output }
  linbuf :            ARRAY [0..255] OF CHAR ;
  srcleng:            integer ;  { length of current source line }
  errmess:            TEXT ;
  errlist:            ARRAY [indextype {0..errlistsize}] OF
		      RECORD   (* ELEMENT DER FEHLERMELDUNGSLISTE *)
			errnr: INTEGER;   (* FEHLERNUMMER *)
			linenr: INTEGER;
			coldig: INTEGER;
		      (* POSITION IN ZEILE, NACH WRITEPOS POSITIONSZIFFER *)
			addstr: integer;  (* pointer to additional string *)
		      END;   (* ERRELEM *)
  errlistuse:         indextype;
  (* ZEIGT AUF DIE LETZTE BELEGTE KOMPONENTE VON ERRLIST *)
  lin,
  sourcelin:          INTEGER;
  (* AKTUELLE ZEILENNUMMER AUS ERRLIST UND QUELLTEXT *)
  startindex,
  stopindex,   (* ANFANG UND ENDE DER AKTUELLEN ZEILE LIN *)
  i:                  indextype;
  (* LAEUFT DURCH ERRLIST *)

  (* LA                   *)
  listall,
  full:               BOOLEAN;   (* "ERRLIST IST UEBERGELAUFEN" *)
  errtext:            TEXT;
  (* ENTFAELLT SPAETER, DA ERRTXT ZUM PROGRAMM GEBUNDEN WIRD *)
  errtxtuse:          integer ;
  errtxt:             ARRAY [1 .. errtxtsize] OF CHAR;

(* FEHLERTEXTE                                                          *)
(* EINTRAEGE IN ERRTXT HABEN FOLGENDE FORM:                             *)
(*  <NUMMER>  <LAENGE>  <FEHLERTEXT>                                    *)
(*     2         1        <LAENGE>    BYTE                              *)
(* <NUMMER> IST DIE FEHLERNUMMER IN INTERNER INTEGER-DARSTELLUNG        *)
(*    (ERMOEGLICHT SCHNELLERE VERARBEITUNG)                             *)
(* <LAENGE>  LAENGE DES FEHLERTEXTES, ERMOEGLICHT SCHNELLERE SUCHE      *)
(* <FEHLERTEXT> HAT DIE LAENGE <LAENGE> UND IST N I C H T DURCH CR      *)
(*              ABGESCHLOSSEN.                                          *)
(* ERRTXT WIRD ENTWEDER VON READERRTEXT BESCHRIEBEN ODER MIT DEM        *)
(* PROGRAMM  GELADEN.                                                   *)
(* Der letzte Eintrag enthaelt nur die Nummer 2^16-1.                   *)

(* Dahinter werden die vom Compiler erzeugten Strings abgelegt, die in  *)
(* den Fehlertext einkopiert werden. Sie haben die Form:                *)
(*  <laenge> <string>                                                   *)

  var  sourcev : string ;
  var  retc : integer ;

$off
procedure strip (var s : filenametype) ;
  var i : integer ;
  begin
    i := filenamesize ;
    while  s[i] = ' '  do i := i - 1 ;
    s[i+1] := chr(0) ;
  end ; { strip }

procedure readfilename ;
  var i : integer ;
  begin
    i := 1 ;
    sourcev := '' ;
    while  not eoln (errmess)  do
      begin
	sourcev := sourcev + errmess^ ;
	get (errmess) ;
	i := i + 1 ;
      end ;
    get (errmess) ;
  end ;
$on


PROCEDURE readerrtext;

  (* LIEST DIE FEHLERTEXTE VON DER DATEI ERRTEXT IN DAS ARRAY ERRTXT, *)
  (* ENTFALLT SPAETER, DA ERRTXT ZUM PROGRAMM GEBUNDEN WIRD.          *)
  (* Format der Datei errtext: Pro Zeile                              *)
  (*  <dezimale Fehlernummer> <Fehlertext>                            *)
  (* Fuehrende Leerzeichen im Fehlertext werden geloescht.            *)
  (* Zeilen, die mit ';', '{' oder '/' beginnen, und Leerzeilen       *)
  (* werden ignoriert.                                                *)

  VAR
    i,   (* DURCHLAEUFT ERRTXT *)
    lin,   (* ZEILENNUMMER VON ERRTEXT *)
    nr,   (* FEHLERNUMMER *)
    lengindex: INTEGER;   (* POSITION DES LAENGENBYTE IN ERRTXT *)

  BEGIN
    i := errtxtuse;
    lin := 1;
    WHILE NOT EOF (errtext)  DO
      IF  eoln (errtext)  or  (errtext^ = '{')  or  (errtext^ = '/')
	  or (errtext^ = '!')  or  (errtext^ = ';') THEN
      BEGIN   (* KOMMENTARZEILE UEBERLESEN *)
	READLN (errtext);
	lin := SUCC (lin);
      END
      ELSE BEGIN   (* LESESCHLEIFE FUER ERRTEXT *)
	READ (errtext, nr);
	errtxt [i] := CHR (nr DIV 256);
	errtxt [i+1] := CHR (nr MOD 256);
	i := i + 2;
	lengindex := i;   (* POSITION DES LAENGENBYTE MERKEN *)
	i := i + 1 ;
	while  errtext^ = ' '  do   get (errtext) ;
	WHILE NOT EOLN (errtext) DO BEGIN
	  READ (errtext, errtxt [i]);
	  i := i+1
	END;   (* TEXT GELESEN *)
	errtxt [lengindex] := CHR (i - lengindex - 1);
	(* LAENGE NACHTRAGEN *)
	GET (errtext);   (* = READLN *)
	lin := lin+1;
      END;   (* UNTIL EOF (ERRTEXT) *)
    errtxt [i] := CHR( 255 );
    errtxt [i + 1] := CHR( 255 );    (* LETZTE ZEICHEN IN ERRTXT (EOF) *)
			 (* Datenstruktur einer Fehlernummer erwartet! *)
    errtxtuse := i + 2 ;
  END;   (* READERRTEXT *)


PROCEDURE readerrmess;

  (* LIEST ERRMES EIN UND SORTIERT DIE FEHLERMELDUNGEN IN ERRLIST EIN *)
  (* BIS EOF (ERRMES) ODER ERRLIST VOLL.                              *)

  VAR
    lin,
    col: INTEGER;   (* AKTUELLE ZEILENNUMMER UND POSITION *)
    enr: integer;   (* aktuelle Fehlernummer *)
    addstring: integer;
    full : boolean; (* indicates errtxt overflow *)


  PROCEDURE meserror (nr: INTEGER);

    BEGIN
      WRITE (output, '!*** ERROR ');
      write (output, 'Internal heap overflow yields loss of information');
      WRITELN (output, ' AT LINE ', lin: 1, ' COL ', col: 1);
    END;


  PROCEDURE insert (col: INTEGER) ;

    (* FUEGT ERRMES@ AN DIE RICHTIGE STELLE (NACH LIN UND POS GEORDNET *)
    (* ODER HINTER INDEX BEI GLOBAL-INSERT) VON ERRLIST                *)
    (* EIN UND BESETZT VOR, WEIST INDEX DIE POSITION IN ERRLIST ZU.    *)
    (* BEI ERRMES@.TAGFD=GLOB MUSS INDEX = INSIND SEIN.                *)


    VAR
      i,
      j: integer;

    BEGIN
      i := errlistuse;
      errlistuse := errlistuse + 1 ;
	WHILE lin < errlist [i].linenr DO
	  i := i - 1 ;
	WHILE (lin = errlist [i].linenr) AND (col < errlist [i].coldig) DO
	  i := i - 1 ;
	i := i + 1 ;   (* I IST JETZT DER RICHTIGE INDEX *)
	FOR  j := errlistuse - 1  DOWNTO i DO
	  errlist [j + 1] := errlist [j];   (* VERSCHIEBEN *)
	WITH errlist [i] DO BEGIN
	  errnr := enr ;
	  linenr := lin ;
	  coldig := col ;
	  addstr := addstring ;
	END;   (* WITH ERRLIST [I] *)
    END;   (* INSERT *)

  BEGIN   (* READERRMES *)
    full := false ;
    lin := 0;
    col := 0;
    WITH errlist [0] DO BEGIN   (* NOTWENDIG FUER INSERT *)
      linenr := - MAXINT;
      coldig := - MAXINT;
    END;
    WHILE EOF (errmess) < (errlistuse < errlistsize) DO BEGIN
      read (errmess, enr, lin, col) ;
      if  eoln (errmess) or full  then  addstring := 0
      else
	begin           (* read additional string *)
	  if  errmess^ = ' '  then get (errmess) ;
	  addstring := errtxtuse ;
	  errtxtuse := errtxtuse + 1 ;
	  while  not eoln (errmess)  and  (errtxtuse < errtxtsize )  do
	   begin
	     read (errmess, errtxt [errtxtuse]) ;
	     errtxtuse := errtxtuse + 1 ;
	   end ;
	  if  (errtxtuse >= errtxtsize)  then
	   begin
	     errtxtuse := errtxtsize-1 ;
	     full := true ;
	     meserror (10) ;
	   end
	  else
	    errtxt [addstring] := chr (errtxtuse - addstring - 1) ;
	end ;           (* additional string read *)
      readln (errmess) ;
      insert(col);
    END;
  END;   (* READERRMES *)


PROCEDURE listline (VAR source : TEXT) ;
  (* SCHREIBT DIE LAUFENDE QUELLTEXTZEILE. *)
  VAR i : INTEGER ;

  BEGIN
    i := 1 ;
    if  listall  then
     begin
       linbuf [0] := ' ' ;
       WRITE (output, '  ');
     end
    else
     begin
       linbuf [0] := chr (ordtab) ;
       WRITE (output, sourcelin : 5, ':', chr (ordtab));
     end ;

    WHILE NOT EOLN (source) AND NOT EOF(SOURCE) DO BEGIN
      WRITE (output, source^);
      if  i <= 255  then
	if  ord(source^) = ordtab  then   linbuf [i] := source^
				   else   linbuf [i] := ' ' ;
      i := i + 1 ;
      GET (source)
    END;
    IF  NOT EOF(SOURCE)  THEN   GET (source);   (* = READLN *)
    if  i > 255  then   srcleng := 255
		 else   srcleng := i ;
    linbuf [srcleng] := ' ' ;   { eoln char }
    WRITELN (output);
    sourcelin := SUCC (sourcelin)
  END;   (* LISTLINE *)


PROCEDURE scansource;
  (* LIEST QUELLTEXT BIS ZUR NAECHSTEN FEHLERHAFTEN ZEILE (EXCL.) ODER EOF *)

  BEGIN
    IF (lin < sourcelin) and (lin <> 0) THEN
      WRITELN (output, '!*** ACTUAL: ', sourcelin: 1, ' ASKED: ', lin: 1);

     WHILE EOF (input) < (sourcelin < lin) DO
       IF listall THEN   (* LA *)
	 listline (input)
       ELSE BEGIN
	 WHILE NOT EOLN (INPUT) AND NOT EOF(INPUT) DO   GET (INPUT) ;
	 IF  NOT EOF(INPUT)  THEN   GET (INPUT) ;   { = READLN (input) }
	 sourcelin := SUCC (sourcelin)
       END
  END;   (* SCANSOURCE *)


PROCEDURE writepos;

  (* GIBT DIE ZEILE MIT DEN FEHLERPOSITIONEN AUS UND ERSETZT IN DER *)
  (* FEHLERLISTE ERRLIST DIE POSITIONEN DURCH DIE POSITIONSZIFFERN. *)

  VAR
    j: INTEGER;
    oldcol: INTEGER;   (* LETZTE AUSGEGEBENE POSITION *)
    colchar: integer { 0 .. 9 };   (* LETZTE POSITIONSZIFFER *)
    colisret: BOOLEAN;
    (* "AUF LETZTER POSITION STEHT EIN RETCHAR UND *)
    (* NOCH KEIN FEHLER AUF DIESER POSITION"       *)
    posline2: BOOLEAN;
    (* "EINE 2. ZEILE IST ERFORDERLICH" *)


  PROCEDURE error (nr: INTEGER);

    BEGIN
      WRITE (output, '!*** ERROR ');
$off    only on error used
      CASE nr OF
	20:
	  WRITE (output, ': NOT GLOBAL');
	21:
	  WRITE (output, 'IN GLOBAL');
	22:
	  WRITE (output, 'SYNTAXERROR AT RET');
	23:
	  WRITE (output, 'IN POS OF RET');
	30:
$on
	  WRITE (output, ': destroying TAB. Position digits may be useless.');
$off
      END;
$on
      WRITELN (output);
    END;


  BEGIN
    oldcol := -1 ;
    colchar := 0;
    colisret := FALSE;
    posline2 := FALSE;
    REPEAT
      WITH errlist [i] DO
	BEGIN
	  IF (coldig > oldcol) OR colisret THEN
	    IF colchar = 9 THEN
	      colchar := 0
	    ELSE
	      colchar := SUCC (colchar);
	  IF coldig > oldcol THEN
	    BEGIN   (* NOT GLOBERR AND (POSDIG <> OLDPOS) *)
	      if  ord (linbuf [coldig]) = ordtab  then   error (30) ;
	      linbuf [coldig] := chr (colchar + ord('0'));
	      oldcol := coldig
	    END ;
	  colisret := FALSE;
	  coldig := colchar;   (* POSITION IN POSITIONSZIFFER AENDERN *)
	END;   (* with errlist [i] *)
      i := i + 1 ;
    UNTIL (errlist [i].linenr <> lin)  OR  (i > errlistuse);

    if  listall  then
      write (output, '!')
    else
      write (output, ' ': 6) ;  { indent because of line number }
    for  j := 0  to  srcleng  do    write (output, linbuf[j]) ;
    WRITELN (output);
  END;   (* WRITEPOS *)


PROCEDURE writemes;
  (* GIBT DIE FEHLERMELDUNGEN DER AKTUELLEN ZEILE LIN AUS. *)

  VAR
    txtind: INTEGER;   (* POSITION DER AKTUELLEN FEHLERMELDUNG IN ERRTXT *)


  FUNCTION findtxt (nr: INTEGER) : INTEGER;

    (* SUCHT IN ERRTXT NACH DEM FEHLER DER FEHLERNUMMER NR, *)
    (* INDEX WIRD DIE POSITION IN ERRTXT ODER -1, WENN NICHT GEFUNDEN       *)

    VAR
      i: INTEGER;

    function num (i:integer) : integer ;
      begin
	num := ORD (errtxt [i]) * 256 + ORD (errtxt [i + 1])
      end ;

    BEGIN
      i := 1;
      WHILE (num(i) <> nr) and (num(i) <> 256*255+255) DO
	i := i + 3 + ORD (errtxt [i + 2]);
      IF (num(i) = nr) THEN
	findtxt := i + 2   (* FINDTXT ZEIGT AUF DIE TEXTLAENGE *)
      ELSE
	findtxt := - 1;
    END;   (* FINDTXT *)


  PROCEDURE putstring (txtind: INTEGER; str : integer);

    (* SCHREIBT EINEN TEXT AUS ERRTXT AUS, ORD (ERRTXT [TXTIND]) IST DIE *)
    (* LAENGE DES TEXTS, DER TEXT BEGINNT BEI TXTIND + 1.                *)

    VAR
      i: INTEGER;
      subpos : integer ; (* position of substitute char in errtxt *)

    BEGIN
      subpos := txtind + ORD (errtxt [txtind]) + 1 ;
      if  str <> 0  then                              { search subchar }
	 FOR  i := txtind + 1  TO  txtind + ORD (errtxt [txtind])  DO
	    if  errtxt [i] = subchar  then   subpos := i ;
      FOR  i := txtind+1  TO  subpos - 1  DO
	WRITE (output, errtxt [i]);
      if  str <> 0  then
       begin
	 FOR  i := str + 1  TO str + ORD (errtxt [str]) DO
	   WRITE (output, errtxt [i]);
	 FOR  i := subpos + 1  TO txtind + ORD (errtxt [txtind]) DO
	   WRITE (output, errtxt [i]);
       end ;
    END;   (* PUTSTRING *)


  BEGIN   (* WRITEMES *)
    i := startindex;
    REPEAT
      WITH errlist [i] DO
	 BEGIN
	   if  listall  then    write (output, '! ') ;
	   if  errnr >= fatallimit  then
	     WRITE (output, fatalat)
	   else if  errnr >= checklimit  then
	     WRITE (output, checkat)
	   else if  errnr >= errorlimit  then
	     WRITE (output, errorat)
	   else if  errnr >= warnglimit  then
	     WRITE (output, warngat)
	   else if  errnr >=  infolimit  then
	     WRITE (output,  infoat)
	   ;
	   write (output, coldig: 2, ': ');
	   txtind := findtxt (errnr );
	   IF txtind = - 1 THEN
	     WRITE (output, errnr: 1)
	   ELSE
	     putstring (txtind, addstr);
	   WRITELN (output);
	 END ;
      i := SUCC (i);
    UNTIL i = stopindex;
  END;   (* WRITEMES *)


BEGIN   (* MAIN *)
  listall := full_listing ;
  if  filexists ((errmess1))  then
   begin
     RESET (errmess, errmess1) ;
     if { ((errmess^ >= 'a') and (errmess^ <= 'z') or (errmess^ = '_') or
	  (errmess^ >= 'A') and (errmess^ <= 'Z') )
	 and } not eof (errmess)  then
      begin                            { read source filename from errmess }
	{ readfilename ; }
	readln (errmess, sourcev) ;
	reset (input, sourcev) ;
      end
     else       { use command line specification for source file }
       RESET (input);
   end
  else
   if  listall  then
     RESET (input);       { aborts, if not exists }

  if  filexists ((errmess1)) or listall  then
  begin
    (* REWRITE (output);  output may be substituted by listing *)
    errlistuse := 0;
    errtxtuse := 1;
    if  filexists ((errtext1))  then
     begin
       RESET (errtext, errtext1);
       readerrtext;
     end
    else
     begin
       retc := openp88 (errtext, errtext1, 20) ;
       if  retc > 0  then   readerrtext
       else   writeln ('File ',errtext1, ' not found.') ;
     end ;

    if  filexists ((errmess1))  then   readerrmess;
    full := errlistuse = errlistsize;
    IF full THEN
      errlistuse := errlistuse - 1 ;   (* SICHERHEITSHALBER *)
    sourcelin := 1;
    i := 1;
    WHILE i <= errlistuse DO BEGIN
      lin := errlist [i].linenr;
      scansource;
      IF EOF (input) THEN BEGIN
	if  listall  then   write (output, '! ') ;
	WRITELN (output, 'Unexpected eof    --- no sourceline ---');
	srcleng := 1 ;
	if  listall  then   linbuf [0] := ' '
		     else   linbuf [0] := chr (ordtab) ;
	linbuf [1] := ' ' ;
      END   (* EOF (SOURCE) *)
      ELSE
	listline (input);
      startindex := i;
      writepos;
      stopindex := i;
      writemes;
    END;
    IF full THEN
      WRITELN (output, '!*** TOO MANY ERRORS');
    IF listall THEN
      WHILE NOT EOF (input) DO
	   listline (input)
  end
  else
    writeln ('File  ', errmess1, '  not found!')
END.
